<?php

namespace NoccyLabs\Plugin\Extension;

use NoccyLabs\Plugin\PluginManager;
use NoccyLabs\Plugin\PluginInstance;
use Symfony\Component\DependencyInjection\ContainerAwareInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\DependencyInjection\Reference;

/**
 *
 *
 *
 */
class SymfonyExtension extends AbstractExtension
{

    protected $app;

    protected $container;

    public function getDefaultName()
    {
        return "symfony";
    }

    public function __construct($app=null, $container=null)
    {
        $this->app = $app;
        $this->container = $container;
    }
    
    public function getApp()
    {
        return $this->app;
    }

    public function getContainer()
    {
        return $this->container;
    }

    public function onLoad()
    {
        $this->plugin->extend("getApp", [ $this, "getApp" ]);
        $this->plugin->extend("getContainer", [ $this, "getContainer" ]);

        if ($serviceConf = $this->plugin->getPath("services.json", true)) {
            $this->registerServices($serviceConf);
        }
        if ($commandConf = $this->plugin->getPath("commands.json", true)) {
            $this->registerCommands($commandConf);
        }
        $inst = $this->plugin->getInstance();
        if ($inst && $inst instanceof ContainerAwareInterface) {
            $inst->setContainer($this->container);
        }
    }

    private function registerServices($filename)
    {
        if (!$this->container) {
            fprintf(STDERR, "Warning: Not registering services from %s since no container is assigned to the extension\n", $filename);
            return;
        }
        $json = file_get_contents($filename);
        $conf = json_decode($json);

        if (@empty($conf->services) && @empty($conf->parameters)) {
            foreach ($conf as $name=>$config) {
                $this->bindService($name, $config);
            }
        } else {
            foreach ((array)@$conf->parameters as $name=>$value) {
                $this->bindParameter($name, $value);
            }
            foreach ((array)@$conf->services as $name=>$config) {
                $this->bindService($name, $config);
            }
        }
    }

    private function bindParameter($name, $value)
    {
        $this->container->setParameter($name, $value);
    }

    private function bindService($name, $config)
    {
        $svc = $this->container->register($name, $config->class);
        if ($arguments = @(array)$config->arguments) {
            foreach ($arguments as $argument) {
                if ($argument[0]=="@") {
                    $argument = new Reference(substr($argument,1));
                }
                $svc = $svc->addArgument($argument);
            }
        }
        if ($calls = @(array)$config->calls) {
            foreach ($calls as $call) {
                $args = (array)$call;
                $method = array_shift($args);
                $args = array_map(function ($argument) {
                    if ($argument[0]=="@") {
                        $argument = new Reference(substr($argument,1));
                    }
                    return $argument;
                }, $args[0]);
                $svc = $svc->addMethodCall($method, $args);
            }
        }
        if ($tags = @(array)$config->tags) {
            foreach ($tags as $tag) {
                if (is_string($tag)) {
                    $svc = $svc->addTag($tag);
                } else {
                    $tagname = null;
                    $tagvals = [];
                    foreach ((array)$tag as $k=>$v) {
                        if ($k == "name") {
                            $tagname = $v;
                        } else {
                            $tagvals[$k] = $v;
                        }
                    }
                    $svc = $svc->addTag($tagname, $tagvals);
                }
            }
        }
    }

    private function registerCommands($filename)
    {
        if (!$this->app) {
            fprintf(STDERR, "Warning: Not registering commands from %s since no app is assigned to the extension\n", $filename);
            return;
        }
        $json = file_get_contents($filename);
        $conf = json_decode($json);
        foreach ($conf as $command=>$def) {
            if (!class_exists($def->class)) {
                throw new \Exception("SymfonyExtension: Class {$def->class} not found for command");
            }
            $inst = new $def->class;
            $this->app->add($inst, $command);
        }
    }

}

