<?php

namespace NoccyLabs\Plugin;

class PluginInstance
{
    /** @var string The plugin name (dot-separated) */
    protected $name;
    /** @var string SemVer version number */
    protected $version;
    /** @var string Language or type of plugin implementation */
    protected $lang;
    /** @var string[] The names of the plugins that require this plugin */
    protected $requiredBy = [];
    /** @var string[] The names and versions of the plugins required by this plugin */
    protected $requires = [];
    /** @var string[] Plugin interfaces implemented by the plugin */
    protected $implements = [];
    /** @var array Extra key-value pairs from the plugin.json file */
    protected $extra = [];
    /** @var string The loader stub if any */
    protected $loaderStub;

    protected $pluginClass;
    /** @var array Autoloaders, like composer. represented as [type,namespace,path] */
    protected $autoloaders = [];

    protected $broken = false;

    protected $loader;

    private $isLoading = false;

    protected $manager;

    protected $instance;

    protected $extensions = [];

    protected $methods = [];

    /**
     *
     *
     * @param AbstractLoader $loader The loader of the plugin distribution
     */
    public function __construct(Loader\AbstractLoader $loader)
    {
        $this->loader = $loader;
        $this->readMetadata();
    }
    
    public function setManager(PluginManager $manager)
    {
        $this->manager = $manager;
        $em = $manager->getExtensionManager();


        // load extensions here, they need the manager
        foreach ($this->extensions as $extension=>$state) {
            if ($state === false) {
                try {
                    $ext = $em->createExtension($this, $extension);
                    if (!$ext) {
                        continue;
                    }
                    $ext->onRegister();
                    $this->extensions[$extension] = $ext;
                } catch (\Exception $e) {
                    fprintf(STDERR, "Warning: The extension %s failed to load for plugin %s: %s\n", $extension, $this->name, $e->getMessage());
                }
            }
        }
    }

    public function hasExtension($extension)
    {
        return array_key_exists($extension, $this->extensions) && $this->extensions[$extension];
    }

    /**
     * Read metadata from the loader
     *
     *
     * 
     */
    protected function readMetadata()
    {
        $manifest = $this->loader->getManifest();

        if (empty($manifest)) {
            throw new PluginException("Invalid manifest");
        }

        $this->name = @$manifest->name;
        $this->version = @$manifest->version;

        $this->lang = @$manifest->lang?:"php";

        $this->loaderStub = @$manifest->stub?:@$manifest->loader;
        $this->pluginClass = @$manifest->class;

        foreach ((array)@$manifest->implements as $implements) {
            $this->implements[] = $implements;
        }

        foreach ((array)@$manifest->requires as $package=>$version) {
            $this->requires[$package] = $version;
        }
        
        foreach ((array)@$manifest->extra as $key=>$value) {
            $this->extra[$key] = $value;
        }

        foreach ((array)@$manifest->autoload as $type=>$autoloaders) {
            if (!in_array($type, [ "psr-0", "psr-4", "stub", "files" ])) {
                continue;
            }
            $this->autoloaders[$type] = array_merge((array)@$this->autoloaders[$type],(array)$autoloaders);
        }

        foreach ((array)@$manifest->extensions as $extension) {
            $this->extensions[$extension] = false;
        }

    }

    /**
     *
     *
     * @return string 
     */
    public function getRoot()
    {
        return $this->loader->getRoot();
    }

    public function getPath($path, $test=false)
    {
        return $this->loader->getPath($path, $test);
    }

    /**
     *
     *
     * @return string 
     */
    public function getLoaderType()
    {
        $type = get_class($this->loader);
        $type = substr($type, strrpos($type,"\\")+1);
        return $type;
    }

    public function getName()
    {
        return $this->name;
    }

    public function getVersion()
    {
        return $this->version;
    }

    public function getInterfaces()
    {
        return $this->implements;
    }

    public function hasInterface($name)
    {
        foreach ($this->implements as $interface) {
            if (fnmatch($name, $interface)) return true;
        }
        return false;
    }

    public function getDependencies()
    {
        return $this->requires;
    }

    public function addDependent(PluginInstance $dependent)
    {
        $this->requiredBy[] = $dependent;
    }

    public function getDependents()
    {
        return $this->requiredBy;
    }

    public function getAllValues()
    {
        return $this->extra;
    }

    public function getValue($name)
    {
        return array_key_exists($name, $this->extra)?$this->extra[$name]:null;
    }

    public function getLanguage()
    {
        return $this->lang;
    }

    public function setBroken($reason)
    {
        $this->broken = $reason;
    }

    public function isBroken()
    {
        return (!!$this->broken);
    }

    public function getBrokenReason()
    {
        return $this->broken;
    }

    public function __call($method, $args)
    {
        if (!array_key_exists($method, $this->methods)) {
            throw new \BadMethodCallException("Bad method call [{$this->name}]->{$method}()");
        }
        return call_user_func($this->methods[$method], $this, ...$args);
    }

    public function extend($method, callable $handler)
    {
        $this->methods[$method] = $handler;
    }

    public function getLoadedExtensions()
    {
        $ret = [];
        foreach ($this->extensions as $extension=>$inst) {
            if ($inst) $ret[] = $extension;
        }
        return $ret;
    }

    public function getExtensions()
    {
        return $this->extensions;
    }

    public function getInstance()
    {
        return $this->instance;
    }

    public function isLoaded()
    {
        return $this->manager->isLoaded($this->name);
    }

    public function load()
    {
        if ($this->manager->isLoaded($this->name)) {
            return;
        }
        
        try {
            $this->loadDependencies();
            $this->loadImplementation();
        } catch (\Throwable $t) {
            throw new PluginManagerException("Unable to load plugin: {$t->getMessage()}", 0, $t);
        }
        $this->manager->setLoaded($this->name, $this->version);
    }

    public function unload()
    {
        if (!$this->manager->isLoaded($this->name)) {
            return;
        }
        
        try {
            $this->unloadImplementation();
        } catch (\Throwable $t) {
            throw new PluginManagerException("Unable to unload plugin: {$t->getMessage()}");
        }

        $this->manager->setLoaded($this->name, false);
    }

    protected function loadDependencies()
    {
        if ($this->isLoading) {
            throw new PluginManagerException("Circular dependency detected when loading {$this->name}");
        }
        $this->isLoading = true;
        foreach ($this->requires as $dependency=>$version) {
            if (!$this->manager->loadPlugin($dependency, $version)) {
                throw new PluginManagerException("Unable to load plugin: Dependency failed");
            }
        }
        $this->isLoading = false;
    }

    protected function loadImplementation()
    {
        $autoloadFromStub = false;
        $definedLoaders = [];

        if (count($this->autoloaders)>0) {
            foreach ($this->autoloaders as $type=>$autoloaders) {
                switch ($type) {
                    case 'psr-0':
                    case 'psr-4':
                        foreach ($autoloaders as $ns=>$path) {
                            $definedLoaders[] = $this->loader->createAutoloader($type,$ns,$path);
                        }
                        break;
                    case 'files':
                        foreach ($autoloaders as $file) {
                            $lfile = $this->loader->getPath($file);
                            require_once $lfile;
                        }
                        break;
                    case 'stub':
                        $autoloadFromStub = true;
                }
            }
        }

        foreach ($definedLoaders as $loader) {
            spl_autoload_register($loader);
        }

        if ($this->loaderStub) {
            $stub = $this->loader->getPath($this->loaderStub);
            $inst = require_once $stub;
            if ($autoloadFromStub) {
                // create autoloader here
            }
            $this->instance = $inst;
        } elseif ($this->pluginClass) {
            if (!class_exists($this->pluginClass)) {
                throw new PluginException("No such class {$this->pluginClass}");
            }
            $this->instance = new $this->pluginClass;
        } else {
            $this->instance = true;
        }
        foreach ($this->extensions as $extension) {
            if ($extension) {
                try {
                    $extension->onLoad();
                } catch (\Exception $e) {
                    throw new PluginException($e->getMessage());
                }
            }
        }

        if ($this->instance) {
            if ($this->instance instanceof PluginAwareInterface) {
                $this->instance->setPluginInstance($this);
                $this->instance->onLoad();
            }
        } else {
            throw new PluginException("No instance resolved for plugin {$this->pluginClass}");
        }

    }

    protected function unloadImplementation()
    {
        if ($this->instance) {
            if ($this->instance instanceof PluginAwareInterface) {
                $this->instance->onUnload();
            }
        }
    }

}
